#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2012  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

'''
Wrapper for the Chemical Identifier Resolver web service at NIH.
'''

import openbabel as ob
import pybel
import sqlite3

from MemoizeDB import MemoizeDB as MDB

from argparse import ArgumentParser, Action
from collections import OrderedDict
from urllib.request import urlopen
from urllib.parse import quote
from urllib.error import HTTPError

import xdg.BaseDirectory
import os.path
import errno



def make_parent_dirs(path, mode=0o777, exist_ok=True):
  '''
  Recursively create all parent directories of the given path.
  '''
  try:
    dname = os.path.dirname(path)
    if dname:
      os.makedirs(dname, mode=mode, exist_ok=exist_ok)
  except OSError as e:
    if e.errno != errno.EEXIST:
      raise e


def strip_inchi_prefix(inchi_key):
  '''
  Remove "InChi=" prefixes from InChi Keys if present.
  '''
  try:
    prefix, inchi_key = inchi_key.split('=', 1)
    return inchi_key
  except (ValueError, AttributeError):
    return inchi_key





# Look up an identifier on CACTUS using an InChi Key and return.
def cir_lookup(identifier, representation):
  '''
  Query the Chemical Identifier Resolver web service with the identifier
  and return the specified representation.

  Return None if no representation was returned by the server.
  '''
  # The CIR URL
  URL = 'http://cactus.nci.nih.gov/chemical/structure/'
  url = "%s%s/%s" % (URL, quote(identifier), quote(representation))
  try:
    f = urlopen(url)
    # From ('Content-Type', 'text/html; charset=ISO-8859-1') header.
    encoding = f.getheader('Content-Type').split('charset=', 1)[1]
    value = f.read().decode(encoding)
    f.close()
  except HTTPError as e:
    if e.getcode() == 404:
      # Return a tuple instead of None to temporarily cache missing values.
      return None
    else:
      raise e

  return value






class CIRError(Exception):
  '''
  Exception raised by fatal ChemicalIdentifierResolver errors.
  '''

  def __init__(self, msg):
    self.msg = msg





# CIR: Chemical Identifier Resolver
class ChemicalIdentifierResolver():
  '''
  Retrieve identifiers from the CACTUS Chemical Identifier Resolver

    http://cactus.nci.nih.gov/chemical/structure/
    http://cactus.nci.nih.gov/chemical/structure/documentation

  The results are cached in an SQLite3 database for future queries. A table is
  created for each representation in ChemicalIdentifierResolver.representations
  and each table uses an InChi Key as a unique identifier, except for the
  "stdinchikey" table, which maps other identifiers to Inchi Keys.

  Representations that are available directly via Open Babel are not included.
  '''

  # The CIR representations and their short descriptions.
  #
  # Not all of the possible representations are included because most of them
  # are directly accessible via Open Babel. There is no point in using a remote
  # service to obtain these, and they would only increase the size of the
  # database.
  #
  # If the full range of representations is required, update the object's
  # representations with the missing entries below.
  REPRESENTATIONS = {
    'stdinchikey' : 'Standard InChIKey',
#     'stdinchi' : 'Standard InChI',
#     'smiles' : 'SMILES',
    'ficts' : 'FICTS Identifier',
    'ficus' : 'FICuS Identifier',
    'uuuuu' : 'uuuuu Identifier',
    'hashisy' : 'Cactvs HASHISY',
#     'sdf' : 'SD File',
    'names' : 'Names',
    'iupac_name' : 'IUPAC Name',
    'cas' : 'CAS Registry Number',
#     'image' : 'GIF Image',
#     'twirl' : 'TwirlyMol (3D)',
#     'mw' : 'Molecular Weight',
#     'formula' : 'Chemical Formula',
#     'h_bond_donor_count' : 'Number of Hydrogen Bond Donors',
#     'h_bond_acceptor_count' : 'Number of Hydrogen Bond Acceptors',
#     'h_bond_center_count' : 'Number of Hydrogen Bond Acceptors and Donors',
#     'rule_of_5_violation_count' : 'Number of Rule of 5 Violations',
#     'rotor_count' : 'Number of Freely Rotatable Bonds',
#     'effective_rotor_count' : 'Number of effectively Rotatable Bonds'
  }

  # The table to map InChi Keys to non-InChi Key inputs.
  INCHI_TABLE = 'INCHI'

  TABLES = set(REPRESENTATIONS.keys())
  TABLES.add(INCHI_TABLE)


  def __init__(self, database=None):
    '''
    database: None, a filepath, ":memory:", or an sqlite3.Connection.
    '''
    if isinstance(database, sqlite3.Connection):
      conn = database
    else:
      if not database:
        cache_dir = xdg.BaseDirectory.save_cache_path('CACTUS')
        database = os.path.join(cache_dir, 'CIR.sqlite3')

      if database != ':memory:':
        # Ensure that the parent directories exist.
        make_parent_dirs(database, exist_ok=True)

      conn = sqlite3.connect(
        database,
        detect_types=sqlite3.PARSE_DECLTYPES|sqlite3.PARSE_COLNAMES,
        isolation_level=None
      )

    glue = {}
    for rep in tuple(self.REPRESENTATIONS):
      if rep == 'stdinchikey':
        def f(xs):
          for x in xs:
            yield x, (strip_inchi_prefix(cir_lookup(x, rep)),)
      else:
        def f(xs):
          for x in xs:
            yield x, (cir_lookup(x, rep),)
      glue[rep] = (f, ((rep, 'TEXT'),), None)

#     glue['names'] = (lambda x: cir_lookup(x, 'names'), (('names', 'TEXT'),), None)

    # Let empty values live for a day.
    self.mdb = MDB(conn, glue, null_ttl=60*60*24)
#     self.mdb.db_clean()

    # Open Babel converter
    self.conv = ob.OBConversion()
    self.conv.SetOutFormat('inchi')
    self.conv.SetOptions('K', self.conv.OUTOPTIONS)

    # Open Babel molecule
    self.mol = ob.OBMol()





  ##############################################################################
  # Logging Functions
  ##############################################################################
  def log(self, msg):
    '''
    Log messages to STDERR.Override this in a subclass for proper logging.
    '''
    sys.stderr.write(msg + "\n")

  def warn(self, msg):
    '''
    Log a warning.
    '''
    self.log("warning: " + msg)

  def die(self, msg):
    '''
    Log an error message, close the database, and raise a CIRError exception.
    '''
    self.log("error: " + msg)
    raise CIRError(msg)
    #sys.exit(1)





  ##############################################################################
  # Open Babel Functions
  ##############################################################################

  def get_inchi_key_from_string(self, string, format):
    '''
    Generate an InChi Key from a string.

    The format must be one of the input formats in pybel.informats.
    '''
    if format not in pybel.informats:
      self.warn('unrecognized input format: %s\n' % format)
      return None

    self.conv.SetInFormat(format)
    self.conv.ReadString(self.mol, string)
    return self.conv.WriteString(self.mol).strip()



  def get_inchi_key_from_file(self, fpath):
    '''
    Generate an InChi Key from a file via Open Babel. The file may be in any
    input format from which Open Babel can generate an InChi Key.

    See pybel.informats for a full list of input formats.

    Return None if it fails.
    '''
    format = self.conv.FormatFromExt(fpath)
    if not format:
      format = self.conv.FormatFromMIME(fpath)
    if not format:
      self.warn('unrecognized format: %s\n' % fpath)
      return None

    self.conv.SetInFormat(format)
    self.conv.ReadFile(self.mol, fpath)
    return self.conv.WriteString(self.mol).strip()



  ##############################################################################
  # CACTUS Functions
  ##############################################################################

  def cir_identifier_to_inchi_key(self, identifier):
    '''
    Return the identifier's InChi Key.

    The InChi Key will be determined by querying the Chemical Identifier
    Resolver with the identifier. The returned key will be cached for future
    requests.

    If the identifier can be recognized by Open Babel, use
    "get_inchi_key_from_string" instead to avoid unnecessary invocations of the
    remote web service..
    '''
    return self.mdb.get('stdinchikey', identifier)



  ##############################################################################
  # Other Functions
  ##############################################################################

  def get(self, rep, identifier):
    '''
    Look up the representation associated with the InChi Key.

    If the representation can not be found in the database, the Chemical
    Identifier Resolver web service will be queried.

    Return None if the representation could not be retrieved.

    '''
    return self.mdb.get(rep, identifier)


  def get_from_file(self, representation, fpath):
    '''
    Generate the InChi Key from a file using "get_inchi_key_from_file" and
    use it to look up the representation with "lookup".

    '''
    inchi_key = self.get_inchi_key_from_file(fpath)
    return self.get(representation, inchi_key)














class DisplayMoreHelp(Action):
  '''
  List all known representations with their descriptions and other information.
  '''
  def __call__(self, parser, namespace, values, option_string=None):
    '''
    Display the additional help message and exit.
    '''

    print('''OUTPUT
  The following Chemical Identifier Resolver output representations are
  available:
''')
    self.display_table('Representation', 'Description', ChemicalIdentifierResolver.REPRESENTATIONS)
    print('''
  See the Chemical Identifier Resolver documentation for more details:

    http://cactus.nci.nih.gov/chemical/structure/documentation

  Note that to avoid redundancy, representations provided directly by Open Babel
  are not included.
''')

    print('''INPUT
  The following Open Babel input formats are available:
''')
    self.display_table('Format', 'Description', pybel.informats)
    print('''
  Note that some of these format might not make sense in a particular context.
  Also note that these only apply to Open Babel input. See the Chemical
  Identifier Resolver documentation for details about which structure
  identifiers it recognizes:

    http://cactus.nci.nih.gov/chemical/structure/documentation
''')

    sys.exit(0)


  def display_table(self, header_l, header_r, rows):
    '''
    Format a two-column table with a left and right header.
    '''

    # The line character to underline the headers.
    c = '-'

    col_l = len(header_l)
    col_r = len(header_r)

    for k, v in rows.items():
      w = len(k)
      if w > col_l:
        col_l = w
      w = len(v)
      if w > col_r:
        col_r = w

    fmt = '    %-' + str(col_l) + 's %s'

    print(fmt % (header_l, header_r))
    print(fmt % (c * col_l, c * col_r))
    for k,v in sorted(rows.items()):
      print(fmt % (k,v))





def parse_args(args=None):
  '''
  Parse command-line arguments.
  '''

  parser = ArgumentParser(description='Query the CACTUS (CADD Group Chemoinformatics Tools and User Services: http://cactus.nci.nih.gov/) Chemical Identifier Resolver using chemical identifier strings or chemical files as input. Files may be in any format from which Open Babel can obtain an InChi Key. Query results are cached in an SQLite3 database.')

  parser.add_argument(
    '--more-help',
    nargs=0,
    action=DisplayMoreHelp,
    help="Display more help including descriptions of Open Babel input formats and Chemical Identifier Resolver output representations."
  )

  parser.add_argument(
    'identifiers',
    #action="append",
    default=[],
    nargs='*',
    metavar='<identifier>',
    help='Identifier strings. By default, these are interpretted as InChi Keys. To use another string format recognized by Open Babel, use the "--format" option. To use an identifier only recognized by the Chemical Structure Identifier, use "--other".'
  )

  parser.add_argument(
    '--format',
    dest="format",
    choices=sorted(pybel.informats),
    help="Select one of the Open Babel input formats for identifier strings passed on the command line. File formats are determined directly from the file and are not affected by this option."
  )

  parser.add_argument(
    '-o', '--other',
    dest='others',
    #action="append",
    default=[],
    nargs='+',
    metavar='<identifier>',
    help="Identifiers recognized by the Chemical Structure Identifier server, e.g. OPSIN-parsable systematic IUPAC names."
  )

  parser.add_argument(
    '-f', '--files',
    dest="files",
    #action="append",
    default=[],
    nargs='+',
    metavar='<filepath>',
    help="Filepaths to files that can be read by Open Babel. InChi Keys will be extracted from these."
  )

  parser.add_argument(
    '-r', '--representation',
    dest="representations",
    #action="append",
    default=[],
    choices=ChemicalIdentifierResolver.REPRESENTATIONS,
    nargs='+',
    help="Specify the output representation(s). See \"-l\" for more information. If this option is not given, all representations will be displayed. Multiple representations may be specified."
  )

  parser.add_argument(
    '-d', '--database',
    action="store",
    dest="database",
    metavar="<database path>",
    help='Set the path to the SQLite3 database of cached results. If this option is not given, the database will be created in "$XDG_CACHE_HOME" if it is set, otherwise it will be created in the current working directory. If the given path is a directory, the database will be created within that directory.'
  )

  return parser.parse_args(args)





def main(args=None):
  '''
  Parse command-line arguments and return the requested data.
  '''

  args = parse_args(args)

  if not args.representations:
    args.representations = sorted(ChemicalIdentifierResolver.REPRESENTATIONS.keys())

#   cir = ChemicalIdentifierResolver(':memory:')
  cir = ChemicalIdentifierResolver(database=args.database)

  # Track input arguments for inclusion with the output later.
  inchi_keys = OrderedDict()

  # Collect InChi Keys for Open Babel identifiers strings.
  for identifier in args.identifiers:
    if args.format:
      inchi_key = cir.get_inchi_key_from_string(identifier, args.format)
      if inchi_key:
        inchi_keys[inchi_key] = identifier
      else:
        cir.warn('failed to generate InChi Key for "%s"' % (identifier))
    # If there's no format then it's already an InChi Key.
    else:
      inchi_key = strip_inchi_prefix(identifier)
      inchi_keys[inchi_key] = None

  # Collect InChi Keys for Open Babel files.
  for fpath in args.files:
    inchi_key = cir.get_inchi_key_from_file(fpath)
    if inchi_key:
        inchi_keys[inchi_key] = 'file: %s' % (fpath)
    else:
      cir.warn('failed to generate InChi Key for "%s"' % (fpath))

  # Collect InChi Keys for Chemical Identifier Resolver identifiers.
  for other in args.others:
    inchi_key = cir.cir_identifier_to_inchi_key(other)
    if inchi_key:
        inchi_keys[inchi_key] = other
    else:
      cir.warn('failed to retrieve InChi Key for "%s"' % (other))

  for inchi_key, source in inchi_keys.items():
    if source:
      print("InChi Key: %s (%s)" % (inchi_key, source) )
    else:
      print("InChi Key: %s" % (inchi_key) )
    for rep in args.representations:
      value, = cir.get(rep, inchi_key)
      if value:
        print("  %s" % cir.REPRESENTATIONS[rep])
        for line in value.strip().split("\n"):
          print("    %s" % line)
        print("")





if __name__ == '__main__':
  try:
    main()
  except (KeyboardInterrupt, BrokenPipeError):
    pass
